#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
#               2009       Gary Burton
# Copyright (C) 2011       Tim G L Lyons
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
The EditPersonRef module provides the EditPersonRef class.
"""

# -------------------------------------------------------------------------
#
# Python modules
#
# -------------------------------------------------------------------------

import pickle

# -------------------------------------------------------------------------
#
# GTK/Gnome modules
#
# -------------------------------------------------------------------------
from gi.repository import Gtk

# -------------------------------------------------------------------------
#
# gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.sgettext
from gramps.gen.display.name import displayer as name_displayer
from .editsecondary import EditSecondary
from .objectentries import PersonEntry
from gramps.gen.lib import NoteType
from ..widgets import MonitoredEntry, PrivacyButton
from .displaytabs import CitationEmbedList, NoteTab
from ..glade import Glade
from ..ddtargets import DdTargets
from gi.repository import Gdk
from gramps.gen.const import URL_MANUAL_SECT1

# -------------------------------------------------------------------------
#
# Constants
#
# -------------------------------------------------------------------------

WIKI_HELP_PAGE = URL_MANUAL_SECT1
WIKI_HELP_SEC = _("Person_Reference_Editor", "manual")


# -------------------------------------------------------------------------
#
# EditPersonRef class
#
# -------------------------------------------------------------------------
class EditPersonRef(EditSecondary):
    """
    Displays a dialog that allows the user to edit a person reference.
    """

    def __init__(self, dbstate, uistate, track, personref, callback):
        """
        Displays the dialog box.

        personref - The person reference that is to be edited
        """
        EditSecondary.__init__(self, dbstate, uistate, track, personref, callback)

    def _local_init(self):
        self.top = Glade()
        self.set_window(
            self.top.toplevel,
            self.top.get_object("title"),
            _("Person Reference Editor"),
        )
        self.setup_configs("interface.person-ref", 600, 350)

        self.person_label = self.top.get_object("person")
        self.share_btn = self.top.get_object("select")
        self.add_del_btn = self.top.get_object("add_del")

    def _setup_fields(self):
        self.person_field = PersonEntry(
            self.dbstate,
            self.uistate,
            self.track,
            self.top.get_object("person"),
            self.top.get_object("person_event_box"),
            self.obj.set_reference_handle,
            self.obj.get_reference_handle,
            self.add_del_btn,
            self.share_btn,
        )

        self.street = MonitoredEntry(
            self.top.get_object("relationship"),
            self.obj.set_relation,
            self.obj.get_relation,
            self.db.readonly,
        )

        self.priv = PrivacyButton(
            self.top.get_object("private"), self.obj, self.db.readonly
        )

    def _connect_signals(self):
        self.define_cancel_button(self.top.get_object("cancel"))
        self.define_ok_button(self.top.get_object("ok"), self.save)
        self.define_help_button(
            self.top.get_object("help"), WIKI_HELP_PAGE, WIKI_HELP_SEC
        )

    def _connect_db_signals(self):
        """
        Connect any signals that need to be connected.
        Called by the init routine of the base class (_EditPrimary).
        """
        self._add_db_signal("person-rebuild", self.close)
        self._add_db_signal("person-delete", self.check_for_close)

    def check_for_close(self, handles):
        """
        Callback method for delete signals.
        If there is a delete signal of the primary object we are editing, the
        editor (and all child windows spawned) should be closed
        """
        if self.obj.ref in handles:
            self.close()

    def _create_tabbed_pages(self):
        """
        Create the notebook tabs and inserts them into the main
        window.
        """

        notebook = Gtk.Notebook()

        self.srcref_list = CitationEmbedList(
            self.dbstate,
            self.uistate,
            self.track,
            self.obj.get_citation_list(),
            "personref_editor_ref_citations",
        )
        self._add_tab(notebook, self.srcref_list)
        self.track_ref_for_deletion("srcref_list")

        self.note_tab = NoteTab(
            self.dbstate,
            self.uistate,
            self.track,
            self.obj.get_note_list(),
            "personref_editor_ref_notes",
            notetype=NoteType.ASSOCIATION,
        )
        self._add_tab(notebook, self.note_tab)
        self.track_ref_for_deletion("note_tab")

        self._setup_notebook_tabs(notebook)
        notebook.show_all()
        self.top.get_object("vbox").pack_start(notebook, True, True, 0)

    def build_menu_names(self, obj):
        return (_("Person Reference"), _("Person Reference Editor"))

    def save(self, *obj):
        """
        Called when the OK button is pressed. Gets data from the
        form and updates the data structure.
        """

        if self.obj.ref:
            if self.callback:
                self.callback(self.obj)
            self.callback = None
            self.close()
        else:
            from ..dialog import ErrorDialog

            ErrorDialog(
                _("No person selected"),
                _("You must either select a person or Cancel the edit"),
                parent=self.uistate.window,
            )



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2005  Donald N. Allingham
# Copyright (C) 2007-2008  Brian G. Matherly
# Copyright (C) 2008       Peter Landgren
# Copyright (C) 2010       Jakim Friant
# Copyright (C) 2012,2017  Paul Franklin
# Copyright (C) 2014       Nick Hall
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Paragraph/Font style editor
"""

# ------------------------------------------------------------------------
#
# Python modules
#
# ------------------------------------------------------------------------
import logging

log = logging.getLogger(".")
import re

# ------------------------------------------------------------------------
#
# GNOME/GTK modules
#
# ------------------------------------------------------------------------
from gi.repository import Gtk, Gdk

# ------------------------------------------------------------------------
#
# Gramps modules
#
# ------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.sgettext
from gramps.gen.plug.docgen import (
    StyleSheet,
    FONT_SERIF,
    FONT_SANS_SERIF,
    PARA_ALIGN_RIGHT,
    PARA_ALIGN_CENTER,
    PARA_ALIGN_LEFT,
    PARA_ALIGN_JUSTIFY,
    ParagraphStyle,
    TableStyle,
    TableCellStyle,
    GraphicsStyle,
)
from ...listmodel import ListModel
from ...managedwindow import ManagedWindow
from ...glade import Glade
from ...dialog import ErrorDialog
from ...display import display_help
from gramps.gen.const import URL_MANUAL_PAGE

WIKI_HELP_PAGE = URL_MANUAL_PAGE + "_-_Settings"


# ------------------------------------------------------------------------
#
# StyleListDisplay class
#
# ------------------------------------------------------------------------
class StyleListDisplay(ManagedWindow):
    """
    Shows the available paragraph/font styles. Allows the user to select,
    add, edit, and delete styles from a StyleSheetList.
    """

    def __init__(self, stylesheetlist, uistate, track, callback=None):
        """
        Create a StyleListDisplay object that displays the styles in the
        StyleSheetList.

        stylesheetlist - styles for editing: a :class:`.StyleSheetList` instance
        callback - task called when an object has been added.
        """

        ManagedWindow.__init__(self, uistate, track, self.__class__, modal=True)
        # the self.window.run() below makes Gtk make it modal, so any change
        # to the previous line's "modal" would require that line to be changed

        self.callback = callback

        self.sheetlist = stylesheetlist

        self.top = Glade(toplevel="styles")
        self.set_window(
            self.top.toplevel, self.top.get_object("title"), _("Document Styles")
        )
        self.setup_configs("interface.stylelistdisplay", 400, 300)
        self.show()

        self.top.connect_signals(
            {
                "on_ok_clicked": self.on_ok_clicked,
                "on_add_clicked": self.on_add_clicked,
                "on_delete_clicked": self.on_delete_clicked,
                "on_button_press": self.on_button_press,
                "on_edit_clicked": self.on_edit_clicked,
                "on_cancel_clicked": self.__cancel,
                "on_help_btn_clicked": lambda x: display_help(
                    WIKI_HELP_PAGE, _("Document_Styles_dialog", "manual")
                ),
                "on_cancel_style_clicked": dummy_callback,
                "on_save_style_clicked": dummy_callback,
                "on_help_btn_style_clicked": dummy_callback,
            }
        )

        self.list = ListModel(
            self.top.get_object("list"),
            [(_("Style"), -1, 10)],
        )
        self.redraw()
        # the self.window.run() makes Gtk make it modal, so any change to that
        # line would require the ManagedWindow.__init__ to be changed also
        self.window.run()
        if self.opened:
            self.close()

    def build_menu_names(self, obj):  # meaningless while it's modal
        """Override :class:`.ManagedWindow` method."""
        return (_("Document Styles"), " ")

    def __cancel(self, obj):
        pass

    def redraw(self):
        """Redraws the list of styles that are currently available"""

        self.list.model.clear()
        self.list.add([_("default")])

        index = 1
        for style in sorted(self.sheetlist.get_style_names()):
            if style == "default":
                continue
            self.list.add([style])
            index += 1

    def on_add_clicked(self, obj):
        """Called when the ADD button is clicked. Invokes the StyleEditor to
        create a new style"""
        style = self.sheetlist.get_style_sheet("default")
        StyleEditor(_("New Style"), style, self)

    def on_ok_clicked(self, obj):
        """Called when the OK button is clicked; Calls the callback task,
        then saves the stylesheet."""
        if self.callback is not None:
            self.callback()
        try:
            self.sheetlist.save()
        except IOError as msg:
            ErrorDialog(_("Error saving stylesheet"), str(msg), parent=self.window)
        except:
            log.error("Failed to save stylesheet", exc_info=True)

    def on_button_press(self, obj, event):
        if event.type == Gdk.EventType.DOUBLE_BUTTON_PRESS and event.button == 1:
            self.on_edit_clicked(obj)

    def on_edit_clicked(self, obj):
        """
        Called when the EDIT button is clicked.
        Calls the StyleEditor to edit the selected style.
        """
        store, node = self.list.selection.get_selected()
        if not node:
            ErrorDialog(
                _("Missing information"), _("Select a style"), parent=self.window
            )
            return

        name = str(self.list.model.get_value(node, 0))
        if name == _("default"):  # the default style cannot be edited
            return
        style = self.sheetlist.get_style_sheet(name)
        StyleEditor(name, style, self)

    def on_delete_clicked(self, obj):
        """Deletes the selected style."""
        store, node = self.list.selection.get_selected()
        if not node:
            ErrorDialog(
                _("Missing information"), _("Select a style"), parent=self.window
            )
            return
        name = str(self.list.model.get_value(node, 0))
        if name == _("default"):  # the default style cannot be removed
            return
        self.sheetlist.delete_style_sheet(name)
        self.redraw()


# ------------------------------------------------------------------------
#
# StyleEditor class
#
# ------------------------------------------------------------------------
class StyleEditor(ManagedWindow):
    """
    Edits the current style definition.
    Presents a dialog allowing the values in the style to be altered.
    """

    def __init__(self, name, style, parent):
        """
        Create the StyleEditor.

        name - name of the style that is to be edited
        style - style object to be edited: a :class:`.StyleSheet` instance
        parent - StyleListDisplay object that called the editor
        """

        ManagedWindow.__init__(
            self, parent.uistate, parent.track, self.__class__, modal=True
        )
        # the self.window.run() below makes Gtk make it modal, so any change
        # to the previous line's "modal" would require that line to be changed

        self.current_style = None
        self.current_name = None

        self.style = StyleSheet(style)
        self.parent = parent
        self.top = Glade(
            toplevel="editor",
            also_load=[
                "adjustment1",
                "adjustment2",
                "adjustment3",
                "adjustment4",
                "adjustment5",
                "adjustment6",
                "adjustment7",
                "adjustment8",
                "adjustment9",
                "adjustment10",
                "adjustment11",
            ],
        )
        self.set_window(
            self.top.toplevel, self.top.get_object("title"), _("Style editor")
        )
        self.setup_configs("interface.styleeditor", 550, 610)
        self.show()

        self.top.connect_signals(
            {
                "on_save_style_clicked": self.on_save_style_clicked,
                "on_cancel_style_clicked": self.__cancel,
                "on_help_btn_style_clicked": lambda x: display_help(
                    WIKI_HELP_PAGE, _("Style_editor_dialog", "manual")
                ),
                "on_cancel_clicked": dummy_callback,
                "on_ok_clicked": dummy_callback,
                "on_add_clicked": dummy_callback,
                "on_delete_clicked": dummy_callback,
                "on_button_press": dummy_callback,
                "on_edit_clicked": dummy_callback,
                "on_help_btn_clicked": dummy_callback,
            }
        )

        self.pname = self.top.get_object("pname")
        self.pdescription = self.top.get_object("pdescription")

        self.notebook = self.top.get_object("notebook1")
        self.vbox = self.top.get_object("column_widths")

        self.line_style = self.top.get_object("line_style")
        line_styles = Gtk.ListStore(int, str)
        line_styles.append([0, "Solid"])
        line_styles.append([1, "Dashed"])
        line_styles.append([2, "Dotted"])
        self.line_style.set_model(line_styles)
        renderer_text = Gtk.CellRendererText()
        self.line_style.pack_start(renderer_text, True)
        self.line_style.add_attribute(renderer_text, "text", 1)

        self.top.get_object("label6").set_text(_("pt", "point size"))

        titles = [(_("Style"), 0, 130)]
        self.plist = ListModel(
            self.top.get_object("ptree"), titles, self.change_display
        )

        for widget_name in ("color", "bgcolor", "line_color", "fill_color"):
            color = self.top.get_object(widget_name)
            label = self.top.get_object(widget_name + "_code")
            color.connect("notify::rgba", self.rgba_changed, label)

        self.top.get_object("style_name").set_text(name)

        def _alphanumeric_sort(iterable):
            """sort the given iterable in the way that humans expect"""
            convert = lambda text: int(text) if text.isdigit() else text
            sort_key = lambda k: [convert(c) for c in re.split("([0-9]+)", k)]
            return sorted(iterable, key=sort_key)

        names = _alphanumeric_sort(self.style.get_paragraph_style_names())
        for p_name in names:
            self.plist.add([p_name], self.style.get_paragraph_style(p_name))
        names = _alphanumeric_sort(self.style.get_table_style_names())
        for t_name in names:
            self.plist.add([t_name], self.style.get_table_style(t_name))
        names = _alphanumeric_sort(self.style.get_cell_style_names())
        for c_name in names:
            self.plist.add([c_name], self.style.get_cell_style(c_name))
        names = _alphanumeric_sort(self.style.get_draw_style_names())
        for d_name in names:
            self.plist.add([d_name], self.style.get_draw_style(d_name))
        self.plist.select_row(0)

        # the self.window.run() makes Gtk make it modal, so any change to that
        # line would require the ManagedWindow.__init__ to be changed also
        self.window.run()
        if self.opened:
            self.close()

    def build_menu_names(self, obj):  # meaningless while it's modal
        """Override :class:`.ManagedWindow` method."""
        return (_("Style editor"), None)

    def __cancel(self, obj):
        pass

    def show_pages(self, show_pages):
        """
        Make the given pages visible.
        """
        for page_num in range(self.notebook.get_n_pages()):
            page = self.notebook.get_nth_page(page_num)
            if page_num in show_pages:
                page.show()
            else:
                page.hide()

    def draw(self):
        """
        Updates the display with the selected style.
        """
        if isinstance(self.current_style, ParagraphStyle):
            self.show_pages([0, 1, 2])
            self.draw_paragraph()
        elif isinstance(self.current_style, TableStyle):
            self.show_pages([0, 3])
            self.draw_table()
        elif isinstance(self.current_style, TableCellStyle):
            self.show_pages([0, 4])
            self.draw_cell()
        elif isinstance(self.current_style, GraphicsStyle):
            self.show_pages([0, 5])
            self.draw_graphics()

    def draw_graphics(self):
        """
        Updates the display with the selected graphics style.
        """
        g = self.current_style
        self.pname.set_text(
            '<span size="larger" weight="bold">%s</span>' % self.current_name
        )
        self.pname.set_use_markup(True)

        descr = g.get_description()
        descr = descr or _("No description available")
        p_style = g.get_paragraph_style()
        if p_style:
            para_note = _("(Embedded style '%s' must be edited separately)")
            descr += "\n\n" + para_note % p_style
        self.pdescription.set_text(descr)

        self.top.get_object("line_style").set_active(g.get_line_style())
        self.top.get_object("line_width").set_value(g.get_line_width())

        line_color = tuple2rgba(g.get_color())
        self.top.get_object("line_color").set_rgba(line_color)
        fill_color = tuple2rgba(g.get_fill_color())
        self.top.get_object("fill_color").set_rgba(fill_color)

        self.top.get_object("shadow").set_active(g.get_shadow())
        self.top.get_object("shadow_space").set_value(g.get_shadow_space())

    def draw_cell(self):
        """
        Updates the display with the selected cell style.
        """
        c = self.current_style
        self.pname.set_text(
            '<span size="larger" weight="bold">%s</span>' % self.current_name
        )
        self.pname.set_use_markup(True)

        descr = c.get_description()
        self.pdescription.set_text(descr or _("No description available"))

        self.top.get_object("cell_lborder").set_active(c.get_left_border())
        self.top.get_object("cell_rborder").set_active(c.get_right_border())
        self.top.get_object("cell_tborder").set_active(c.get_top_border())
        self.top.get_object("cell_bborder").set_active(c.get_bottom_border())
        self.top.get_object("cell_padding").set_value(c.get_padding())

    def draw_table(self):
        """
        Updates the display with the selected table style.
        """
        t = self.current_style
        self.pname.set_text(
            '<span size="larger" weight="bold">%s</span>' % self.current_name
        )
        self.pname.set_use_markup(True)

        descr = t.get_description()
        self.pdescription.set_text(descr or _("No description available"))

        self.top.get_object("table_width").set_value(t.get_width())

        self.column = []
        for widget in self.vbox.get_children():
            self.vbox.remove(widget)

        for i in range(t.get_columns()):
            hbox = Gtk.Box()
            label = Gtk.Label(label=_("Column %d:") % (i + 1))
            hbox.pack_start(label, False, False, 6)
            spin = Gtk.SpinButton()
            spin.set_range(0, 100)
            spin.set_increments(1, 10)
            spin.set_numeric(True)
            spin.set_value(t.get_column_width(i))
            self.column.append(spin)
            hbox.pack_start(spin, False, False, 6)
            hbox.pack_start(Gtk.Label(label="%"), False, False, 6)
            hbox.show_all()
            self.vbox.pack_start(hbox, False, False, 3)

    def draw_paragraph(self):
        """
        Updates the display with the selected paragraph style.
        """
        p = self.current_style
        self.pname.set_text(
            '<span size="larger" weight="bold">%s</span>' % self.current_name
        )
        self.pname.set_use_markup(True)

        descr = p.get_description()
        self.pdescription.set_text(descr or _("No description available"))

        font = p.get_font()
        self.top.get_object("size").set_value(font.get_size())
        if font.get_type_face() == FONT_SERIF:
            self.top.get_object("roman").set_active(1)
        else:
            self.top.get_object("swiss").set_active(1)
        self.top.get_object("bold").set_active(font.get_bold())
        self.top.get_object("italic").set_active(font.get_italic())
        self.top.get_object("underline").set_active(font.get_underline())
        if p.get_alignment() == PARA_ALIGN_LEFT:
            self.top.get_object("lalign").set_active(1)
        elif p.get_alignment() == PARA_ALIGN_RIGHT:
            self.top.get_object("ralign").set_active(1)
        elif p.get_alignment() == PARA_ALIGN_CENTER:
            self.top.get_object("calign").set_active(1)
        else:
            self.top.get_object("jalign").set_active(1)
        self.top.get_object("rmargin").set_value(p.get_right_margin())
        self.top.get_object("lmargin").set_value(p.get_left_margin())
        self.top.get_object("pad").set_value(p.get_padding())
        self.top.get_object("tmargin").set_value(p.get_top_margin())
        self.top.get_object("bmargin").set_value(p.get_bottom_margin())
        self.top.get_object("indent").set_value(p.get_first_indent())
        self.top.get_object("tborder").set_active(p.get_top_border())
        self.top.get_object("lborder").set_active(p.get_left_border())
        self.top.get_object("rborder").set_active(p.get_right_border())
        self.top.get_object("bborder").set_active(p.get_bottom_border())

        color = tuple2rgba(font.get_color())
        self.top.get_object("color").set_rgba(color)
        bg_color = tuple2rgba(p.get_background_color())
        self.top.get_object("bgcolor").set_rgba(bg_color)

    def rgba_changed(self, color, name, label):
        """
        Called to set the color code when a color is changed.
        """
        rgb = rgba2tuple(color.get_rgba())
        label.set_text("#%02X%02X%02X" % rgb)

    def save(self):
        """
        Saves the current style displayed on the dialog.
        """
        if isinstance(self.current_style, ParagraphStyle):
            self.save_paragraph()
        elif isinstance(self.current_style, TableStyle):
            self.save_table()
        elif isinstance(self.current_style, TableCellStyle):
            self.save_cell()
        elif isinstance(self.current_style, GraphicsStyle):
            self.save_graphics()

    def save_graphics(self):
        """
        Saves the current graphics style displayed on the dialog.
        """
        g = self.current_style
        g.set_line_style(self.top.get_object("line_style").get_active())
        g.set_line_width(self.top.get_object("line_width").get_value())
        line_color = self.top.get_object("line_color").get_rgba()
        g.set_color(rgba2tuple(line_color))
        fill_color = self.top.get_object("fill_color").get_rgba()
        g.set_fill_color(rgba2tuple(fill_color))
        shadow = self.top.get_object("shadow").get_active()
        shadow_space = self.top.get_object("shadow_space").get_value()
        g.set_shadow(shadow, shadow_space)

        self.style.add_draw_style(self.current_name, self.current_style)

    def save_cell(self):
        """
        Saves the current cell style displayed on the dialog.
        """
        c = self.current_style
        c.set_left_border(self.top.get_object("cell_lborder").get_active())
        c.set_right_border(self.top.get_object("cell_rborder").get_active())
        c.set_top_border(self.top.get_object("cell_tborder").get_active())
        c.set_bottom_border(self.top.get_object("cell_bborder").get_active())
        c.set_padding(self.top.get_object("cell_padding").get_value())

        self.style.add_cell_style(self.current_name, self.current_style)

    def save_table(self):
        """
        Saves the current table style displayed on the dialog.
        """
        t = self.current_style
        t.set_width(self.top.get_object("table_width").get_value_as_int())
        for i in range(t.get_columns()):
            t.set_column_width(i, self.column[i].get_value_as_int())

        self.style.add_table_style(self.current_name, self.current_style)

    def save_paragraph(self):
        """
        Saves the current paragraph style displayed on the dialog.
        """
        p = self.current_style
        font = p.get_font()
        font.set_size(self.top.get_object("size").get_value_as_int())

        if self.top.get_object("roman").get_active():
            font.set_type_face(FONT_SERIF)
        else:
            font.set_type_face(FONT_SANS_SERIF)

        font.set_bold(self.top.get_object("bold").get_active())
        font.set_italic(self.top.get_object("italic").get_active())
        font.set_underline(self.top.get_object("underline").get_active())
        if self.top.get_object("lalign").get_active():
            p.set_alignment(PARA_ALIGN_LEFT)
        elif self.top.get_object("ralign").get_active():
            p.set_alignment(PARA_ALIGN_RIGHT)
        elif self.top.get_object("calign").get_active():
            p.set_alignment(PARA_ALIGN_CENTER)
        else:
            p.set_alignment(PARA_ALIGN_JUSTIFY)

        p.set_right_margin(self.top.get_object("rmargin").get_value())
        p.set_left_margin(self.top.get_object("lmargin").get_value())
        p.set_top_margin(self.top.get_object("tmargin").get_value())
        p.set_bottom_margin(self.top.get_object("bmargin").get_value())
        p.set_padding(self.top.get_object("pad").get_value())
        p.set_first_indent(self.top.get_object("indent").get_value())
        p.set_top_border(self.top.get_object("tborder").get_active())
        p.set_left_border(self.top.get_object("lborder").get_active())
        p.set_right_border(self.top.get_object("rborder").get_active())
        p.set_bottom_border(self.top.get_object("bborder").get_active())

        color = self.top.get_object("color").get_rgba()
        font.set_color(rgba2tuple(color))
        bg_color = self.top.get_object("bgcolor").get_rgba()
        p.set_background_color(rgba2tuple(bg_color))

        self.style.add_paragraph_style(self.current_name, self.current_style)

    def on_save_style_clicked(self, obj):
        """
        Saves the current style sheet and causes the parent to be updated with
        the changes.
        """
        name = str(self.top.get_object("style_name").get_text())

        self.save()
        self.style.set_name(name)
        self.parent.sheetlist.set_style_sheet(name, self.style)
        self.parent.redraw()

    def change_display(self, obj):
        """
        Called when the paragraph selection has been changed. Saves the
        old paragraph, then draws the newly selected paragraph.
        """
        # Don't save until current_name is defined
        # If it's defined, save under the current paragraph name
        if self.current_name:
            self.save()
        # Then change to new paragraph
        objs = self.plist.get_selected_objects()
        store, node = self.plist.get_selected()
        self.current_name = store.get_value(node, 0)
        self.current_style = objs[0]
        self.draw()


def tuple2rgba(rgb):
    """
    Convert a tuple containing 8-bit RGB values into a Gdk.RGBA.
    """
    rgba = Gdk.RGBA()
    rgba.red = rgb[0] / 0xFF
    rgba.green = rgb[1] / 0xFF
    rgba.blue = rgb[2] / 0xFF
    return rgba


def rgba2tuple(rgba):
    """
    Convert a Gdk.RGBA into a tuple containing 8-bit RGB values.
    """
    return (int(rgba.red * 0xFF), int(rgba.green * 0xFF), int(rgba.blue * 0xFF))


def dummy_callback(obj):
    """Dummy callback to satisfy gtkbuilder on connect of signals.
    There are two widgets in the glade file, although only one is needed,
    the signals of the other must be connected too
    """
    pass



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2009 Benny Malengier
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
from gramps.gen.plug._pluginreg import newplugin, STABLE, IMPORT
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

MODULE_VERSION = "6.0"

# ------------------------------------------------------------------------
#
# Comma _Separated Values Spreadsheet (CSV)
#
# ------------------------------------------------------------------------

_mime_type = "text/x-comma-separated-values"  # CSV Document
_mime_type_rfc_4180 = "text/csv"  # CSV Document   See rfc4180 for mime type
plg = newplugin()
plg.id = "im_csv"
plg.name = _("Comma Separated Values Spreadsheet (CSV)")
plg.description = _("Import data from CSV files")
plg.version = "1.0"
plg.gramps_target_version = MODULE_VERSION
plg.status = STABLE
plg.fname = "importcsv.py"
plg.ptype = IMPORT
plg.import_function = "importData"
plg.extension = "csv"

# ------------------------------------------------------------------------
#
# GEDCOM
#
# ------------------------------------------------------------------------

plg = newplugin()
plg.id = "im_ged"
plg.name = _("GEDCOM")
plg.description = _(
    "GEDCOM is used to transfer data between genealogy programs. "
    "Most genealogy software will accept a GEDCOM file as input."
)
plg.version = "1.0"
plg.gramps_target_version = MODULE_VERSION
plg.status = STABLE
plg.fname = "importgedcom.py"
plg.ptype = IMPORT
plg.import_function = "importData"
plg.extension = "ged"

# ------------------------------------------------------------------------
#
# Geneweb
#
# ------------------------------------------------------------------------

plg = newplugin()
plg.id = "im_geneweb"
plg.name = _("GeneWeb")
plg.description = _("Import data from GeneWeb files")
plg.version = "1.0"
plg.gramps_target_version = MODULE_VERSION
plg.status = STABLE
plg.fname = "importgeneweb.py"
plg.ptype = IMPORT
plg.import_function = "importData"
plg.extension = "gw"

# ------------------------------------------------------------------------
#
# Gramps package (portable XML)
#
# ------------------------------------------------------------------------

plg = newplugin()
plg.id = "im_gpkg"
plg.name = _("Gramps package (portable XML)")
plg.description = _(
    "Import data from a Gramps package (an archived XML "
    "Family Tree together with the media object files)."
)
plg.version = "1.0"
plg.gramps_target_version = MODULE_VERSION
plg.status = STABLE
plg.fname = "importgpkg.py"
plg.ptype = IMPORT
plg.import_function = "impData"
plg.extension = "gpkg"

# ------------------------------------------------------------------------
#
# Gramps XML database
#
# ------------------------------------------------------------------------

plg = newplugin()
plg.id = "im_gramps"
plg.name = _("Gramps XML Family Tree")
plg.description = _(
    "The Gramps XML format is a text "
    "version of a Family Tree. It is "
    "read-write compatible with the "
    "present Gramps database format."
)
plg.version = "1.0"
plg.gramps_target_version = MODULE_VERSION
plg.status = STABLE
plg.fname = "importxml.py"
plg.ptype = IMPORT
plg.import_function = "importData"
plg.extension = "gramps"

# ------------------------------------------------------------------------
#
# GRDB database
#
# ------------------------------------------------------------------------

plg = newplugin()
plg.id = "im_grdb"
plg.name = _("Gramps 2.x database")
plg.description = _("Import data from Gramps 2.x database files")
plg.version = "1.0"
plg.gramps_target_version = MODULE_VERSION
plg.status = STABLE
plg.fname = "importgrdb.py"
plg.ptype = IMPORT
plg.import_function = "importData"
plg.extension = "grdb"

# ------------------------------------------------------------------------
#
# Pro-Gen Files
#
# ------------------------------------------------------------------------

plg = newplugin()
plg.id = "im_progen"
plg.name = _("Pro-Gen")
plg.description = _("Import data from Pro-Gen files")
plg.version = "1.0"
plg.gramps_target_version = MODULE_VERSION
plg.status = STABLE
plg.fname = "importprogen.py"
plg.ptype = IMPORT
plg.import_function = "_importData"
plg.extension = "def"

# ------------------------------------------------------------------------
#
# vCard
#
# ------------------------------------------------------------------------

plg = newplugin()
plg.id = "im_vcard"
plg.name = _("vCard")
plg.description = _("Import data from vCard files")
plg.version = "1.0"
plg.gramps_target_version = MODULE_VERSION
plg.status = STABLE
plg.fname = "importvcard.py"
plg.ptype = IMPORT
plg.import_function = "importData"
plg.extension = "vcf"



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2004-2005  Donald N. Allingham
# Copyright (C) 2010       Jakim Friant
# Copyright (C) 2011       Paul Franklin
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# Written by Alex Roitman

"""
General option handling, including saving and parsing.
"""

# -------------------------------------------------------------------------
#
# Python modules
#
# -------------------------------------------------------------------------
import os
import sys

# -------------------------------------------------------------------------
#
# SAX interface
#
# -------------------------------------------------------------------------
from xml.sax import make_parser, handler, SAXParseException
from xml.sax.saxutils import quoteattr

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from ..const import GRAMPS_LOCALE as glocale
from ..plug import BasePluginManager
from ..utils.cast import get_type_converter
from .menu import Menu

PLUGMAN = BasePluginManager.get_instance()

_ = glocale.translation.gettext


# -------------------------------------------------------------------------
#
# List of options for a single module
#
# -------------------------------------------------------------------------
class OptionList:
    """
    Implements a set of options to parse and store for a given module.
    """

    def __init__(self):
        self.options = {}

    def set_options(self, options):
        """
        Set the whole bunch of options for the OptionList.

        :param options: list of options to set.
        :type options: list
        """
        self.options = options

    def get_options(self):
        """
        Return the whole bunch of  options for the OptionList.

        :returns: list of options
        :rtype: list
        """
        return self.options

    def set_option(self, name, value):
        """
        Set a particular option in the OptionList.

        :param name: name of the option to set.
        :type name: str
        :param value: value of the option to set.
        :type value: str
        """
        self.options[name] = value

    def remove_option(self, name):
        """
        Remove a particular option from the OptionList.

        :param name: name of the option to remove.
        :type name: str
        """
        if name in self.options:
            del self.options[name]

    def get_option(self, name):
        """
        Return the value of a particular option in the OptionList.

        :param name: name of the option to retrieve
        :type name: str
        :returns: value associated with the passed option
        :rtype: str
        """
        return self.options.get(name, None)


# -------------------------------------------------------------------------
#
# Collection of option lists
#
# -------------------------------------------------------------------------
class OptionListCollection:
    """
    Implements a collection of option lists.
    """

    def __init__(self, filename):
        """
        Create an OptionListCollection instance from the list defined
        in the specified file.

        :param filename: XML file that contains option definitions
        :type filename: str
        """

        self.filename = os.path.expanduser(filename)
        self.option_list_map = {}
        self.docgen_names = PLUGMAN.get_docgen_names()
        self.init_common()
        self.parse()

    def init_common(self):
        """
        Stub function for common initialization
        """

    def get_option_list_map(self):
        """
        Return the map of module names to option lists.

        :returns: Returns the map of module names to option lists.
        :rtype: dictionary
        """
        return self.option_list_map

    def get_option_list(self, name):
        """
        Return the option_list associated with the module name

        :param name: name associated with the desired module.
        :type name: str
        :returns: returns the option list associated with the name,
                  or None of no such option exists
        :rtype: str
        """
        return self.option_list_map.get(name, None)

    def get_module_names(self):
        """
        Return a list of all the module names in the OptionListCollection

        :returns: returns the list of module names
        :rtype: list
        """
        return list(self.option_list_map.keys())

    def set_option_list(self, name, option_list):
        """
        Add or replaces an option_list in the OptionListCollection.

        :param name: name associated with the module to add or replace.
        :type name: str
        :param option_list: list of options
        :type option_list: OptionList
        """
        self.option_list_map[name] = option_list

    def write_common(self, filename):
        """
        Stub function for common options. Overridden by reports.
        """

    def write_module_common(self, filename, option_list):
        """
        Stub function for common options. Overridden by reports.
        """

    def save(self):
        """
        Saves the current OptionListCollection to the associated file.
        """
        with open(self.filename, "w", encoding="utf-8") as file:
            file.write('<?xml version="1.0" encoding="utf-8"?>\n')
            file.write("<options>\n")

            self.write_common(file)

            for module_name in sorted(self.get_module_names()):  # enable a diff
                option_list = self.get_option_list(module_name)
                module_docgen_opts = {}
                for docgen_name in self.docgen_names:
                    module_docgen_opts[docgen_name] = []
                file.write(f"<module name={quoteattr(module_name)}>\n")
                options = option_list.get_options()
                for option_name in sorted(options.keys()):  # enable a diff
                    option_data = options[option_name]
                    if isinstance(option_data, (list, tuple)):
                        if option_data and option_data[0] in self.docgen_names:
                            module_docgen_opts[option_data[0]].append(
                                (option_name, option_data[1])
                            )
                        else:
                            file.write(
                                f"  <option name={quoteattr(option_name)} "
                                f'value="" length="{len(option_data)}">\n'
                            )
                            for list_index, list_data in enumerate(option_data):
                                file.write(
                                    f'    <listitem number="{list_index}" '
                                    f"value={quoteattr(str(list_data))}/>\n"
                                )
                            file.write("  </option>\n")
                    else:
                        file.write(
                            f"  <option name={quoteattr(option_name)} "
                            f"value={quoteattr(str(option_data))}/>\n"
                        )
                for docgen_name in self.docgen_names:
                    if module_docgen_opts[docgen_name]:
                        for data in module_docgen_opts[docgen_name]:
                            file.write(
                                f"  <docgen-option docgen={quoteattr(docgen_name)} "
                                f"name={quoteattr(data[0])} "
                                f"value={quoteattr(str(data[1]))}/>\n"
                            )
                self.write_module_common(file, option_list)

                file.write("</module>\n")

            file.write("</options>\n")

    def parse(self):
        """
        Loads the OptionList from the associated file, if it exists.
        """
        try:
            if os.path.isfile(self.filename):
                parser = make_parser()
                parser.setContentHandler(OptionParser(self))
                parser.parse(self.filename)
        except (IOError, OSError, SAXParseException):
            pass


# -------------------------------------------------------------------------
#
# OptionParser
#
# -------------------------------------------------------------------------
class OptionParser(handler.ContentHandler):
    """
    SAX parsing class for the OptionListCollection XML file.
    """

    def __init__(self, collection):
        """
        Create a OptionParser class that populates the passed collection.

        collection:   OptionListCollection to be loaded from the file.
        """
        handler.ContentHandler.__init__(self)
        self.collection = collection

        self.mname = None
        self.option_list = None
        self.oname = None
        self.odict = None
        self.an_o = None
        self.list_class = OptionList

    def startElement(self, name, attrs):
        """
        Overridden class that handles the start of a XML element
        """
        tag = name
        if tag in ("report", "module"):
            self.mname = attrs["name"]
            self.option_list = self.list_class()
            self.odict = {}
        elif tag == "option":
            self.oname = attrs["name"]
            if "length" in attrs:
                self.an_o = []
            else:
                self.an_o = attrs["value"]
        elif tag == "listitem":
            self.an_o.append(attrs["value"])

    def endElement(self, name):
        "Overridden class that handles the end of a XML element"
        tag = name
        if tag == "option":
            self.odict[self.oname] = self.an_o
        elif tag in ("report", "module"):
            self.option_list.set_options(self.odict)
            self.collection.set_option_list(self.mname, self.option_list)


# -------------------------------------------------------------------------
#
# Class handling options for plugins
#
# -------------------------------------------------------------------------
class OptionHandler:
    """
    Implements handling of the options for the plugins.
    """

    def __init__(self, module_name, options_dict, person_id=None):
        self.module_name = module_name
        self.default_options_dict = options_dict.copy()
        self.options_dict = options_dict

        # Retrieve our options from whole collection
        self.init_subclass()
        self.option_list_collection = self.collection_class(self.filename)
        self.init_common()
        self.saved_option_list = self.option_list_collection.get_option_list(
            module_name
        )
        self.person_id = person_id

        # Whatever was found should override the defaults
        if self.saved_option_list:
            self.set_options()
        else:
            # If nothing was found, set up the option list
            self.saved_option_list = self.list_class()
            self.option_list_collection.set_option_list(
                module_name, self.saved_option_list
            )

    def init_subclass(self):
        """
        Initialize option subclasses
        """
        self.collection_class = OptionListCollection
        self.list_class = OptionList
        self.filename = None

    def init_common(self):
        """
        Stub method to initialize common options, for derived classes
        """

    def set_options(self):
        """
        Set options to be used in this plugin according to the passed
        options dictionary.

        Dictionary values are all strings, since they were read from XML.
        Here we need to convert them to the needed types. We use default
        values to determine the type.
        """
        # First we set options_dict values based on the saved options
        options = self.saved_option_list.get_options()
        bad_opts = []
        for option_name, option_data in options.items():
            if option_name not in self.options_dict:
                bad_opts.append(option_name)
                continue
            try:
                converter = get_type_converter(self.options_dict[option_name])
                self.options_dict[option_name] = converter(option_data)
            except ValueError:
                pass
            except TypeError:
                pass

        docgen_names = self.option_list_collection.docgen_names
        for option_name in bad_opts:
            option_data = options[option_name]
            if not (
                isinstance(option_data, list)
                and option_data
                and option_data[0] in docgen_names
            ):
                print(
                    _(
                        "Option '%(opt_name)s' is present in %(file)s\n"
                        "  but is not known to the module.  Ignoring..."
                    )
                    % {
                        "opt_name": option_name,
                        "file": self.option_list_collection.filename,
                    },
                    file=sys.stderr,
                )
            options.pop(option_name)

        # Then we set common options from whatever was found
        self.set_common_options()

    def set_common_options(self):
        """
        Stub method to set common options, for derived classes
        """

    def save_options(self):
        """
        Saves options to file.

        We need to only store non-default options. Therefore, we remove all
        options whose values are the defaults prior to saving.
        """

        # First we save options from options_dict
        for option_name, option_data in self.options_dict.items():
            if option_data == self.default_options_dict[option_name]:
                self.saved_option_list.remove_option(option_name)
            else:
                self.saved_option_list.set_option(
                    option_name, self.options_dict[option_name]
                )

        # Handle common options
        self.save_common_options()

        # Finally, save the whole collection into file
        self.option_list_collection.save()

    def save_common_options(self):
        """
        Stub method to save common options, for derived classes
        """

    def get_person_id(self):
        """
        Return the person id
        """
        return self.person_id

    def set_person_id(self, val):
        """
        Set the person id
        """
        self.person_id = val


# ------------------------------------------------------------------------
#
# Base Options class
#
# ------------------------------------------------------------------------
class Options:
    """
    Defines options and provides handling interface.

    This is a base Options class for the modules. All modules, options
    classes should derive from it.
    """

    def __init__(self, name, person_id=None):
        """
        Initialize the class, performing usual house-keeping tasks.
        Subclasses MUST call this in their __init__() method.

        Modules that need custom options need to override this method.
        Two dictionaries allow the addition of custom options:

            self.options_dict
                This is a dictionary whose keys are option names
                and values are the default option values.

            self.options_help
                This is a dictionary whose keys are option names
                and values are 3- or 4- lists or tuples:
                    ('=example','Short description',VALUES,DO_PREPEND)
                The VALUES is either a single string (in that case
                the DO_PREPEND does not matter) or a list/tuple of
                strings to list. In that case, if DO_PREPEND evaluates
                as True then each string will be preneded with the ordinal
                number when help is printed on the command line.

        .. note:: Both dictionaries must have identical keys.
        """
        self.name = name
        self.person_id = person_id
        self.options_dict = {}
        self.options_help = {}
        self.handler = None

    def load_previous_values(self):
        """
        Modifies all options to have the value they were last used as.
        Call this function after all options have been added.
        """
        self.handler = OptionHandler(self.name, self.options_dict, self.person_id)

    def add_user_options(self):
        """
        Set up UI controls (widgets) for the options specific for this module.

        This method MUST be overridden by modules that define new options.

        .. note:: To really have any effect besides looking pretty, each widget
                  set up here must be also parsed in the
                  :meth:`parse_user_options` method below.
        """

    def parse_user_options(self):
        """
        Parses UI controls (widgets) for the options specific for this module.

        This method MUST be overridden by modules that define new options.

        After obtaining values from the widgets, they MUST be used to set the
        appropriate options_dict values. Otherwise the values will not have
        any user-visible effect.

        .. note:: Any widget parsed here MUST be defined and added to the dialog
                  in the :meth:`add_user_options` method above.
        """


# ------------------------------------------------------------------------
#
# MenuOptions class
#
# ------------------------------------------------------------------------
class MenuOptions:
    """
    **Introduction**

    A MenuOptions is used to implement the necessary functions for adding
    options to a menu.
    """

    def __init__(self):
        self.menu = Menu()

        # Fill options_dict with report/tool defaults:
        self.options_dict = {}
        self.options_help = {}
        self.add_menu_options(self.menu)
        for name in self.menu.get_all_option_names():
            option = self.menu.get_option_by_name(name)
            self.options_dict[name] = option.get_value()
            self.options_help[name] = ["", option.get_help()]

    def make_default_style(self, default_style):
        """
        This function is currently required by some reports.
        """

    def add_menu_options(self, menu):
        """
        Add the user defined options to the menu.

        :param menu: A menu class for the options to belong to.
        :type menu: Menu
        :return: nothing
        """
        raise NotImplementedError

    def add_menu_option(self, category, name, option):
        """
        Add a single option to the menu.
        """
        self.menu.add_option(category, name, option)
        self.options_dict[name] = option.get_value()
        self.options_help[name] = ["", option.get_help()]

    def add_user_options(self):
        """
        Generic method to add user options to the menu.
        """
        for category in self.menu.get_categories():
            for name in self.menu.get_option_names(category):
                option = self.menu.get_option(category, name)

                # override option default with xml-saved value:
                if name in self.options_dict:
                    option.set_value(self.options_dict[name])

    def parse_user_options(self):
        """
        Load the changed values into the saved options.
        """
        for name in self.menu.get_all_option_names():
            option = self.menu.get_option_by_name(name)
            self.options_dict[name] = option.get_value()



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2006  Donald N. Allingham
# Copyright (C) 2009       Benny Malengier
# Copyright (C) 2010       Nick Hall
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.utils.lru import LRU
from gramps.gen.config import config


class BaseModel:
    # LRU cache size
    _CACHE_SIZE = config.get("interface.treemodel-cache-size")

    def __init__(self):
        self.lru_data = LRU(BaseModel._CACHE_SIZE)
        self.lru_path = LRU(BaseModel._CACHE_SIZE)

    def destroy(self):
        """
        Destroy the items in memory.
        """
        self.lru_data = None
        self.lru_path = None

    def clear_cache(self, handle=None):
        """
        Clear the LRU cache. Always clear lru_path, because paths may have
        changed.
        """
        if handle:
            if handle in self.lru_data:
                del self.lru_data[handle]
        else:
            self.lru_data.clear()
        # Invalidates all paths
        self.lru_path.clear()

    def get_cached_value(self, handle, col):
        """
        Get the value of a "col". col may be a number (position in a model)
        or a name (special value used by view).
        """
        if handle in self.lru_data and col in self.lru_data[handle]:
            # print("hit", handle, col)
            return (True, self.lru_data[handle][col])
        # print("MISS", handle, col)
        return (False, None)

    def set_cached_value(self, handle, col, data):
        """
        Set the data associated with handle + col.
        """
        if not self._in_build:
            if self.lru_data.count > 0:
                if handle not in self.lru_data:
                    self.lru_data[handle] = {}
                self.lru_data[handle][col] = data

    ## Cached Path's for TreeView:
    def get_cached_path(self, handle):
        """
        Saves the Gtk iter path.
        """
        if handle in self.lru_path:
            return (True, self.lru_path[handle])
        return (False, None)

    def set_cached_path(self, handle, path):
        """
        Set the Gtk iter path value.
        """
        if not self._in_build:
            self.lru_path[handle] = path

    def clear_path_cache(self):
        """
        Clear path cache for all.
        """
        self.lru_path.clear()



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright(C) 2014  Bastien Jacquet
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
from gramps.gen.const import GRAMPS_LOCALE as glocale

"""
GtkWidget showing a box for interactive-search in Gtk.TreeView
"""

# -------------------------------------------------------------------------
#
# Python modules
#
# -------------------------------------------------------------------------
import logging

_LOG = logging.getLogger(".widgets.interactivesearch")

# -------------------------------------------------------------------------
#
# GTK modules
#
# -------------------------------------------------------------------------
from gi.repository import Gtk, Gdk, GLib

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from ..utils import match_primary_mask, get_display_size

# -------------------------------------------------------------------------
#
# InteractiveSearchBox class
#
# -------------------------------------------------------------------------


class InteractiveSearchBox:
    """
    Mainly adapted from gtktreeview.c
    """

    _SEARCH_DIALOG_TIMEOUT = 5000
    _SEARCH_DIALOG_LAUNCH_TIMEOUT = 150

    def __init__(self, treeview):
        self._treeview = treeview
        self._search_window = None
        self._search_entry = None
        self._search_entry_changed_id = 0
        self.__disable_popdown = False
        self._entry_flush_timeout = None
        self._entry_launchsearch_timeout = None
        self.__selected_search_result = 0
        # Disable builtin interactive search by intercepting CTRL-F instead.
        # self._treeview.connect('start-interactive-search',
        #                       self.start_interactive_search)

    def treeview_keypress(self, obj, event):
        """
        function handling keypresses from the treeview
        for the typeahead find capabilities
        """
        if not Gdk.keyval_to_unicode(event.keyval):
            return False
        if self._key_cancels_search(event.keyval):
            return False
        self.ensure_interactive_directory()

        # Make a copy of the current text
        old_text = self._search_entry.get_text()

        popup_menu_id = self._search_entry.connect("popup-menu", lambda x: True)

        # Move the entry off screen
        width, height = get_display_size(self._treeview)
        self._search_window.move(width + 1, height + 1)
        self._search_window.show()

        # Send the event to the window.  If the preedit_changed signal is
        # emitted during this event, we will set self.__imcontext_changed
        new_event = Gdk.Event.copy(event)
        new_event.window = self._search_window.get_window()
        self._search_window.realize()
        self.__imcontext_changed = False
        retval = self._search_window.event(new_event)
        self._search_window.hide()

        self._search_entry.disconnect(popup_menu_id)

        # Intercept CTRL+F keybinding because Gtk do not allow to _replace_ it.
        if match_primary_mask(event.state) and event.keyval in [Gdk.KEY_f, Gdk.KEY_F]:
            self.__imcontext_changed = True
            # self.real_start_interactive_search(event.get_device(), True)

        # We check to make sure that the entry tried to handle the text,
        # and that the text has changed.
        new_text = self._search_entry.get_text()
        text_modified = old_text != new_text
        if self.__imcontext_changed or (  # we're in a preedit
            retval and text_modified
        ):  # ...or the text was modified
            self.real_start_interactive_search(event.get_device(), False)
            self._treeview.grab_focus()
            return True
        else:
            self._search_entry.set_text("")
            return False

    def _preedit_changed(self, im_context, tree_view):
        self.__imcontext_changed = 1
        if self._entry_flush_timeout:
            GLib.source_remove(self._entry_flush_timeout)
            self._entry_flush_timeout = GLib.timeout_add(
                self._SEARCH_DIALOG_TIMEOUT, self.cb_entry_flush_timeout
            )

    def ensure_interactive_directory(self):
        toplevel = self._treeview.get_toplevel()
        if self._search_window:
            if toplevel.has_group():
                toplevel.get_group().add_window(self._search_window)
            elif self._search_window.has_group():
                self._search_window.get_group().remove_window(self._search_window)
            return

        self._search_window = Gtk.Window(type=Gtk.WindowType.POPUP)
        if toplevel.has_group():
            toplevel.get_group().add_window(self._search_window)
        self._search_window.set_type_hint(Gdk.WindowTypeHint.UTILITY)
        self._search_window.set_modal(True)
        self._search_window.connect("delete-event", self._delete_event)
        self._search_window.connect("key-press-event", self._key_press_event)
        self._search_window.connect("button-press-event", self._button_press_event)
        self._search_window.connect("scroll-event", self._scroll_event)
        frame = Gtk.Frame()
        frame.set_shadow_type(Gtk.ShadowType.ETCHED_IN)
        frame.show()
        self._search_window.add(frame)

        vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        vbox.show()
        frame.add(vbox)
        vbox.set_border_width(3)

        """ add entry """
        self._search_entry = Gtk.SearchEntry()
        self._search_entry.show()
        self._search_entry.connect("populate-popup", self._disable_popdown)
        self._search_entry.connect("activate", self._activate)
        self._search_entry.connect("preedit-changed", self._preedit_changed)

        vbox.add(self._search_entry)
        self._search_entry.realize()

    def real_start_interactive_search(self, device, keybinding):
        """
        Pops up the interactive search entry.  If keybinding is TRUE then
        the user started this by typing the start_interactive_search
        keybinding. Otherwise, it came from just typing
        """
        if self._search_window.get_visible():
            return True
        self.ensure_interactive_directory()
        if keybinding:
            self._search_entry.set_text("")
        self._position_func()
        self._search_window.show()
        if self._search_entry_changed_id == 0:
            self._search_entry_changed_id = self._search_entry.connect(
                "changed", self.delayed_changed
            )

        # Grab focus without selecting all the text
        self._search_entry.grab_focus()
        self._search_entry.set_position(-1)
        # send focus-in event
        event = Gdk.Event()
        event.type = Gdk.EventType.FOCUS_CHANGE
        event.focus_change.in_ = True
        event.focus_change.window = self._search_window.get_window()
        self._search_entry.emit("focus-in-event", event)
        # search first matching iter
        self.delayed_changed(self._search_entry)
        # uncomment when deleting delayed_changed
        # self.search_init(self._search_entry)
        return True

    def cb_entry_flush_timeout(self):
        event = Gdk.Event()
        event.type = Gdk.EventType.FOCUS_CHANGE
        event.focus_change.in_ = True
        event.focus_change.window = self._treeview.get_window()
        self._dialog_hide(event)
        self._entry_flush_timeout = 0
        return False

    def delayed_changed(self, obj):
        """
        This permits to start the search only a short delay after last keypress
        This becomes useless with Gtk 3.10 Gtk.SearchEntry, which has a
        'search-changed' signal.
        """
        # renew flush timeout
        self._renew_flush_timeout()
        # renew search timeout
        if self._entry_launchsearch_timeout:
            GLib.source_remove(self._entry_launchsearch_timeout)
        self._entry_launchsearch_timeout = GLib.timeout_add(
            self._SEARCH_DIALOG_LAUNCH_TIMEOUT, self.search_init
        )

    def search_init(self):
        """
        This is the function performing the search
        """
        self._entry_launchsearch_timeout = 0
        text = self._search_entry.get_text()
        if not text:
            return

        model = self._treeview.get_model()
        if not model:
            return
        selection = self._treeview.get_selection()
        # disable flush timeout while searching
        if self._entry_flush_timeout:
            GLib.source_remove(self._entry_flush_timeout)
            self._entry_flush_timeout = 0
        # search
        # cursor_path = self._treeview.get_cursor()[0]
        # model.get_iter(cursor_path)
        start_iter = model.get_iter_first()
        self.search_iter(selection, start_iter, text, 0, 1)
        self.__selected_search_result = 1
        # renew flush timeout
        self._renew_flush_timeout()

    def _renew_flush_timeout(self):
        if self._entry_flush_timeout:
            GLib.source_remove(self._entry_flush_timeout)
        self._entry_flush_timeout = GLib.timeout_add(
            self._SEARCH_DIALOG_TIMEOUT, self.cb_entry_flush_timeout
        )

    def _move(self, up=False):
        text = self._search_entry.get_text()
        if not text:
            return

        if up and self.__selected_search_result == 1:
            return False

        model = self._treeview.get_model()
        selection = self._treeview.get_selection()
        # disable flush timeout while searching
        if self._entry_flush_timeout:
            GLib.source_remove(self._entry_flush_timeout)
            self._entry_flush_timeout = 0
        # search
        start_count = self.__selected_search_result + (-1 if up else 1)
        start_iter = model.get_iter_first()
        found_iter = self.search_iter(selection, start_iter, text, 0, start_count)
        if found_iter:
            self.__selected_search_result += -1 if up else 1
            return True
        else:
            # Return to old iter
            self.search_iter(
                selection, start_iter, text, 0, self.__selected_search_result
            )
            return False
        # renew flush timeout
        self._renew_flush_timeout()
        return

    def _activate(self, obj):
        self.cb_entry_flush_timeout()
        # If we have a row selected and it's the cursor row, we activate
        # the row XXX

    #         if self._cursor_node and \
    #             self._cursor_node.set_flag(Gtk.GTK_RBNODE_IS_SELECTED):
    #             path = _gtk_tree_path_new_from_rbtree(
    #                            tree_view->priv->cursor_tree,
    #                            tree_view->priv->cursor_node)
    #             gtk_tree_view_row_activated(tree_view, path,
    #                                         tree_view->priv->focus_column)

    def _button_press_event(self, obj, event):
        if not obj:
            return
        # keyb_device = event.device
        event = Gdk.Event()
        event.type = Gdk.EventType.FOCUS_CHANGE
        event.focus_change.in_ = True
        event.focus_change.window = self._treeview.get_window()
        self._dialog_hide(event)

    def _disable_popdown(self, obj, menu):
        self.__disable_popdown = 1
        menu.connect("hide", self._enable_popdown)

    def _enable_popdown(self, obj):
        self._timeout_enable_popdown = GLib.timeout_add(
            self._SEARCH_DIALOG_TIMEOUT, self._real_search_enable_popdown
        )

    def _real_search_enable_popdown(self):
        self.__disable_popdown = 0

    def _delete_event(self, obj, event):
        if not obj:
            return
        self._dialog_hide(None)

    def _scroll_event(self, obj, event):
        retval = False
        if event.direction == Gdk.ScrollDirection.UP:
            self._move(True)
            retval = True
        elif event.direction == Gdk.ScrollDirection.DOWN:
            self._move(False)
            retval = True
        if retval:
            self._renew_flush_timeout()

    def _key_cancels_search(self, keyval):
        return keyval in [
            Gdk.KEY_Escape,
            Gdk.KEY_Tab,
            Gdk.KEY_KP_Tab,
            Gdk.KEY_ISO_Left_Tab,
        ]

    def _key_press_event(self, widget, event):
        retval = False
        # close window and cancel the search
        if self._key_cancels_search(event.keyval):
            self.cb_entry_flush_timeout()
            return True
        # Launch search
        if event.keyval in [Gdk.KEY_Return, Gdk.KEY_KP_Enter]:
            if self._entry_launchsearch_timeout:
                GLib.source_remove(self._entry_launchsearch_timeout)
                self._entry_launchsearch_timeout = 0
            self.search_init()
            retval = True

        default_accel = widget.get_modifier_mask(Gdk.ModifierIntent.PRIMARY_ACCELERATOR)
        # select previous matching iter
        if (event.keyval in [Gdk.KEY_Up, Gdk.KEY_KP_Up]) or (
            (
                (event.state & (default_accel | Gdk.ModifierType.SHIFT_MASK))
                == (default_accel | Gdk.ModifierType.SHIFT_MASK)
            )
            and (event.keyval in [Gdk.KEY_g, Gdk.KEY_G])
        ):
            if not self._move(True):
                widget.error_bell()
            retval = True

        # select next matching iter
        if (event.keyval in [Gdk.KEY_Down, Gdk.KEY_KP_Down]) or (
            (
                (event.state & (default_accel | Gdk.ModifierType.SHIFT_MASK))
                == (default_accel)
            )
            and (event.keyval in [Gdk.KEY_g, Gdk.KEY_G])
        ):
            if not self._move(False):
                widget.error_bell()
            retval = True

        # renew the flush timeout
        if retval:
            self._renew_flush_timeout()
        return retval

    def _dialog_hide(self, event):
        if self.__disable_popdown:
            return
        if self._search_entry_changed_id:
            self._search_entry.disconnect(self._search_entry_changed_id)
            self._search_entry_changed_id = 0
        if self._entry_flush_timeout:
            GLib.source_remove(self._entry_flush_timeout)
            self._entry_flush_timeout = 0
        if self._entry_launchsearch_timeout:
            GLib.source_remove(self._entry_launchsearch_timeout)
            self._entry_launchsearch_timeout = 0
        if self._search_window.get_visible():
            # send focus-in event
            self._search_entry.emit("focus-out-event", event)
            self._search_window.hide()
            self._search_entry.set_text("")
            self._treeview.emit("focus-in-event", event)
        self.__selected_search_result = 0

    def _position_func(self, userdata=None):
        tree_window = self._treeview.get_window()
        s_width, s_height = get_display_size(tree_window)

        self._search_window.realize()
        ret, tree_x, tree_y = tree_window.get_origin()
        tree_width = tree_window.get_width()
        tree_height = tree_window.get_height()
        _, requisition = self._search_window.get_preferred_size()

        if tree_x + tree_width > s_width:
            x = s_width - requisition.width
        elif tree_x + tree_width - requisition.width < 0:
            x = 0
        else:
            x = tree_x + tree_width - requisition.width

        if tree_y + tree_height + requisition.height > s_height:
            y = s_height - requisition.height
        elif tree_y + tree_height < 0:  # isn't really possible ...
            y = 0
        else:
            y = tree_y + tree_height

        self._search_window.move(x, y)

    def search_iter_slow(self, selection, cur_iter, text, count, n):
        """
        Standard row-by-row search through all rows
        Should work for both List/Tree models
        Both expanded and collapsed rows are searched.
        """
        model = self._treeview.get_model()
        search_column = self._treeview.get_search_column()
        is_tree = not (model.get_flags() & Gtk.TreeModelFlags.LIST_ONLY)
        while True:
            if not cur_iter:  # can happen on empty list
                return False
            if self.search_equal_func(model, search_column, text, cur_iter):
                count += 1
                if count == n:
                    found_path = model.get_path(cur_iter)
                    self._treeview.expand_to_path(found_path)
                    self._treeview.scroll_to_cell(found_path, None, 1, 0.5, 0)
                    selection.select_path(found_path)
                    self._treeview.set_cursor(found_path)
                    return True

            if is_tree and model.iter_has_child(cur_iter):
                cur_iter = model.iter_children(cur_iter)
            else:
                done = False
                while True:  # search iter of next row
                    next_iter = model.iter_next(cur_iter)
                    if next_iter:
                        cur_iter = next_iter
                        done = True
                    else:
                        cur_iter = model.iter_parent(cur_iter)
                        if not cur_iter:
                            # we've run out of tree, done with this func
                            return False
                    if done:
                        break
        return False

    @staticmethod
    def search_equal_func(model, search_column, text, cur_iter):
        value = model.get_value(cur_iter, search_column)
        key1 = value.lower()
        key2 = text.lower()
        return key1.startswith(key2)

    def search_iter(self, selection, cur_iter, text, count, n):
        model = self._treeview.get_model()
        is_listonly = model.get_flags() & Gtk.TreeModelFlags.LIST_ONLY
        if is_listonly and hasattr(model, "node_map"):
            return self.search_iter_sorted_column_flat(
                selection, cur_iter, text, count, n
            )
        else:
            return self.search_iter_slow(selection, cur_iter, text, count, n)

    def search_iter_sorted_column_flat(self, selection, cur_iter, text, count, n):
        """
        Search among the currently set search-column for a cell starting with
        text
        It assumes that this column is currently sorted, and as
        a LIST_ONLY view it therefore contains index2hndl = model.node_map._index2hndl
        which is a _sorted_ list of (sortkey, handle) tuples
        """
        model = self._treeview.get_model()
        search_column = self._treeview.get_search_column()
        is_tree = not (model.get_flags() & Gtk.TreeModelFlags.LIST_ONLY)

        # If there is a sort_key index, let's use it
        if not is_tree and hasattr(model, "node_map"):
            import bisect

            index2hndl = model.node_map._index2hndl

            # create lookup key from the appropriate sort_func
            # TODO: explicitely announce the data->sortkey func in models
            # sort_key = model.sort_func(text)
            sort_key = glocale.sort_key(text.lower())
            srtkey_hndl = (sort_key, "")
            lo_bound = 0  # model.get_path(cur_iter)
            found_index = bisect.bisect_left(index2hndl, srtkey_hndl, lo=lo_bound)
            # if insert position is at tail, no match
            if found_index == len(index2hndl):
                return False
            srt_key, hndl = index2hndl[found_index]
            # Check if insert position match for real
            # (as insert position might not start with the text)
            if not model[found_index][search_column].lower().startswith(text.lower()):
                return False
            found_path = Gtk.TreePath((model.node_map.real_path(found_index),))
            self._treeview.scroll_to_cell(found_path, None, 1, 0.5, 0)
            selection.select_path(found_path)
            self._treeview.set_cursor(found_path)
            return True
        return False



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2006  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# Standard python modules
#
# -------------------------------------------------------------------------
import sys
import html
import logging
import unicodedata

_LOG = logging.getLogger(".dialog")

# -------------------------------------------------------------------------
#
# GNOME/GTK+ modules
#
# -------------------------------------------------------------------------
from gi.repository import GObject
from gi.repository import Gtk
from gi.repository import Gdk
from gi.repository import GdkPixbuf

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext
from gramps.gen.const import ICON as ICON_PATH, URL_BUGHOME
from gramps.gen.config import config
from gramps.gen.constfunc import is_quartz
from .glade import Glade
from .display import display_url

try:
    ICON = GdkPixbuf.Pixbuf.new_from_file(ICON_PATH)
except:
    ICON = None


class SaveDialog:
    def __init__(self, msg1, msg2, task1, task2, parent=None):
        self.xml = Glade(toplevel="savedialog")

        self.top = self.xml.toplevel
        self.top.set_icon(ICON)
        self.top.set_title("%s - Gramps" % msg1)

        self.dontask = self.xml.get_object("dontask")
        self.task1 = task1
        self.task2 = task2

        label1 = self.xml.get_object("sd_label1")
        label1.set_text('<span weight="bold" size="larger">%s</span>' % msg1)
        label1.set_use_markup(True)

        label2 = self.xml.get_object("sd_label2")
        label2.set_text(msg2)
        label2.set_use_markup(True)
        if parent:
            self.top.set_transient_for(parent)
            parent_modal = parent.get_modal()
            if parent_modal:
                parent.set_modal(False)
        self.top.show()
        response = self.top.run()
        if response == Gtk.ResponseType.NO:
            self.task1()
        elif response == Gtk.ResponseType.YES:
            self.task2()

        config.set("interface.dont-ask", self.dontask.get_active())
        self.top.destroy()
        if parent and parent_modal:
            parent.set_modal(True)


class QuestionDialog:
    def __init__(self, msg1, msg2, label, task, parent=None):
        self.xml = Glade(toplevel="questiondialog")

        self.top = self.xml.toplevel
        self.top.set_icon(ICON)
        self.top.set_title("%s - Gramps" % msg1)

        label1 = self.xml.get_object("qd_label1")
        label1.set_text(
            '<span weight="bold" size="larger">%s</span>' % html.escape(msg1)
        )
        label1.set_use_markup(True)

        label2 = self.xml.get_object("qd_label2")
        label2.set_text(msg2)
        label2.set_use_markup(True)

        self.xml.get_object("okbutton").set_label(label)

        if parent:
            self.top.set_transient_for(parent)
            parent_modal = parent.get_modal()
            if parent_modal:
                parent.set_modal(False)
        self.top.show()
        response = self.top.run()
        self.top.destroy()
        if parent and parent_modal:
            parent.set_modal(True)
        if response == Gtk.ResponseType.ACCEPT:
            task()


def on_activate_link(label, uri):
    # see aboutdialog.py _show_url()
    display_url(uri)
    return True


class QuestionDialog2:
    def __init__(self, msg1, msg2, label_msg1, label_msg2, parent=None):
        self.xml = Glade(toplevel="questiondialog")

        self.top = self.xml.toplevel
        self.top.set_icon(ICON)
        self.top.set_title("%s - Gramps" % msg1)

        label1 = self.xml.get_object("qd_label1")
        label1.set_text(
            '<span weight="bold" size="larger">%s</span>' % html.escape(msg1)
        )
        label1.set_use_markup(True)

        label2 = self.xml.get_object("qd_label2")
        # see https://github.com/emesene/emesene/issues/723
        label2.connect("activate-link", on_activate_link)
        label2.set_text(msg2)
        label2.set_use_markup(True)

        self.xml.get_object("okbutton").set_label(label_msg1)
        self.xml.get_object("okbutton").set_use_underline(True)
        self.xml.get_object("no").set_label(label_msg2)
        self.xml.get_object("no").set_use_underline(True)

        self.parent = parent
        if parent:
            self.top.set_transient_for(parent)
            self.parent_modal = parent.get_modal()
            if self.parent_modal:
                parent.set_modal(False)
        self.top.show()

    def run(self):
        response = self.top.run()
        self.top.destroy()
        if self.parent and self.parent_modal:
            self.parent.set_modal(True)
        return response == Gtk.ResponseType.ACCEPT


class QuestionDialog3:
    """Like QuestionDialog2, but includes cancel button,
    returns: 1  (okbutton, label_msg1)
             0  (no button label_msg2)
             -1 (cancel button)
    """

    def __init__(self, msg1, msg2, label_msg1, label_msg2, parent=None):
        self.xml = Glade(toplevel="questiondialog")

        self.top = self.xml.toplevel
        self.top.set_icon(ICON)
        self.top.set_title("%s - Gramps" % msg1)

        label1 = self.xml.get_object("qd_label1")
        label1.set_text(
            '<span weight="bold" size="larger">%s</span>' % html.escape(msg1)
        )
        label1.set_use_markup(True)

        label2 = self.xml.get_object("qd_label2")
        # see https://github.com/emesene/emesene/issues/723
        label2.connect("activate-link", on_activate_link)
        label2.set_text(msg2)
        label2.set_use_markup(True)

        self.xml.get_object("okbutton").set_label(label_msg1)
        self.xml.get_object("okbutton").set_use_underline(True)
        self.xml.get_object("no").set_label(label_msg2)
        self.xml.get_object("no").set_use_underline(True)
        self.xml.get_object("cancelbutton").show()

        self.parent = parent
        if parent:
            self.top.set_transient_for(parent)
            self.parent_modal = parent.get_modal()
            if self.parent_modal:
                parent.set_modal(False)
        self.top.show()

    def run(self):
        response = self.top.run()
        self.top.destroy()
        if self.parent and self.parent_modal:
            self.parent.set_modal(True)
        return (
            -1
            if response == Gtk.ResponseType.DELETE_EVENT
            else response == Gtk.ResponseType.ACCEPT
        )


class OptionDialog:
    def __init__(self, msg1, msg2, btnmsg1, task1, btnmsg2, task2, parent=None):
        self.xml = Glade(toplevel="optiondialog")

        self.top = self.xml.toplevel
        self.top.set_icon(ICON)
        self.top.set_title("%s - Gramps" % msg1)

        label1 = self.xml.get_object("od_label1")
        label1.set_text('<span weight="bold" size="larger">%s</span>' % msg1)
        label1.set_use_markup(True)

        label2 = self.xml.get_object("od_label2")
        label2.set_text(msg2)
        label2.set_use_markup(True)

        self.xml.get_object("option1").set_label(btnmsg1)
        self.xml.get_object("option2").set_label(btnmsg2)
        if parent:
            self.top.set_transient_for(parent)
            parent_modal = parent.get_modal()
            if parent_modal:
                parent.set_modal(False)
        self.top.show()
        self.response = self.top.run()
        if self.response == Gtk.ResponseType.NO:
            if task1:
                task1()
        else:
            if task2:
                task2()
        self.top.destroy()
        if parent and parent_modal:
            parent.set_modal(True)

    def get_response(self):
        return self.response


class ErrorDialog(Gtk.MessageDialog):
    def __init__(self, msg1, msg2="", parent=None):
        Gtk.MessageDialog.__init__(
            self, transient_for=parent, modal=True, message_type=Gtk.MessageType.ERROR
        )
        self.add_button(_("_Close"), Gtk.ResponseType.CLOSE)
        self.set_markup('<span weight="bold" size="larger">%s</span>' % str(msg1))
        self.format_secondary_text(msg2)
        self.set_icon(ICON)
        self.set_title("%s - Gramps" % str(msg1))
        if parent:
            parent_modal = parent.get_modal()
            if parent_modal:
                parent.set_modal(False)
        self.show()
        self.run()
        self.destroy()
        if parent and parent_modal:
            parent.set_modal(True)


class RunDatabaseRepair(ErrorDialog):
    def __init__(self, msg, parent=None):
        ErrorDialog.__init__(
            # These exact strings are also in cli/user.py -- keep them in sync
            self,
            _("Error detected in database"),
            _(
                "Gramps has detected an error in the database. This can "
                'usually be resolved by running the "Check and Repair Database" '
                "tool.\n\nIf this problem continues to exist after running this "
                "tool, please file a bug report at "
                "%(gramps_bugtracker_url)s\n\n"
            )
            % {"gramps_bugtracker_url": URL_BUGHOME}
            + msg,
            parent,
        )


class DBErrorDialog(ErrorDialog):
    def __init__(self, msg, parent=None):
        ErrorDialog.__init__(
            # These exact strings are also in cli/user.py -- keep them in sync
            self,
            _("Low level database corruption detected"),
            _(
                "Gramps has detected a problem in the underlying "
                "database. This can sometimes be repaired from "
                "the Family Tree Manager. Select the database and "
                "click on the Repair button"
            )
            + "\n\n"
            + msg,
            parent,
        )


class WarningDialog(Gtk.MessageDialog):
    def __init__(self, msg1, msg2="", parent=None):
        Gtk.MessageDialog.__init__(
            self, transient_for=parent, modal=True, message_type=Gtk.MessageType.WARNING
        )
        self.add_button(_("_Close"), Gtk.ResponseType.CLOSE)
        self.set_markup('<span weight="bold" size="larger">%s</span>' % msg1)
        self.format_secondary_markup(msg2)
        # FIXME: Hyper-links in the secondary text display as underlined text,
        # but clicking on the link fails with
        # GtkWarning: Unable to show 'http://www.gramps-project.org/wiki/index.php?title=How_to_make_a_backup': Operation not supported
        # self.connect('activate-link'... fails with
        # <WarningDialog object at 0x4880300 (GtkMessageDialog at 0x5686010)>: unknown signal name: activate-link
        self.set_icon(ICON)
        self.set_title("%s - Gramps" % msg1)
        if parent:
            parent_modal = parent.get_modal()
            if parent_modal:
                parent.set_modal(False)
        self.show()
        self.run()
        self.destroy()
        if parent and parent_modal:
            parent.set_modal(True)


class OkDialog(Gtk.MessageDialog):
    def __init__(self, msg1, msg2="", parent=None):
        Gtk.MessageDialog.__init__(
            self, transient_for=parent, modal=True, message_type=Gtk.MessageType.INFO
        )
        self.add_button(_("_Close"), Gtk.ResponseType.CLOSE)
        self.set_markup('<span weight="bold" size="larger">%s</span>' % msg1)
        self.format_secondary_text(msg2)
        self.set_icon(ICON)
        self.set_title("%s - Gramps" % msg1)
        if parent:
            parent_modal = parent.get_modal()
            if parent_modal:
                parent.set_modal(False)
        self.show()
        self.run()
        self.destroy()
        if parent and parent_modal:
            parent.set_modal(True)


class InfoDialog:
    """
    Non modal dialog to show selectable info in a scrolled window
    """

    def __init__(self, msg1, infotext, parent=None, monospaced=False):
        self.xml = Glade(toplevel="infodialog")

        self.top = self.xml.toplevel
        self.top.set_icon(ICON)
        self.top.set_title("%s - Gramps" % msg1)

        label = self.xml.get_object("toplabel")
        label.set_text('<span weight="bold" size="larger">%s</span>' % msg1)
        label.set_use_markup(True)

        infoview = self.xml.get_object("infoview")
        infobuffer = Gtk.TextBuffer()
        infoview.set_buffer(infobuffer)

        if isinstance(infotext, str):
            infobuffer.set_text(infotext)
        else:
            for item in infotext:
                enditer = infobuffer.get_end_iter()
                if isinstance(item, str):
                    infobuffer.insert(enditer, item + "\n")
                elif isinstance(item, list):
                    grid = Gtk.Grid()
                    grid.set_margin_start(6)
                    grid.set_margin_end(6)
                    grid.set_column_spacing(12)
                    if unicodedata.bidirectional(item[0][0][0]) == "R":
                        grid.set_direction(Gtk.TextDirection.RTL)
                    for offset_y, row in enumerate(item):
                        for offset_x, col in enumerate(row):
                            cell = Gtk.Label(col)
                            cell.set_halign(Gtk.Align.END)
                            grid.attach(cell, offset_x, offset_y, 1, 1)
                    grid.show_all()
                    anchor = infobuffer.create_child_anchor(enditer)
                    infoview.add_child_at_anchor(grid, anchor)
                    enditer = infobuffer.get_end_iter()
                    infobuffer.insert(enditer, "\n")

        if monospaced:
            startiter, enditer = infobuffer.get_bounds()
            tag = infobuffer.create_tag(family="Monospace")
            infobuffer.apply_tag(tag, startiter, enditer)

        if parent:
            self.top.set_transient_for(parent)
        self.top.connect("response", self.destroy)
        self.top.show()

    def destroy(self, dialog, response_id):
        # no matter how it finishes, destroy dialog
        dialog.destroy()


class MissingMediaDialog:
    def __init__(self, msg1, msg2, task1, task2, task3, parent=None):
        self.xml = Glade(toplevel="missmediadialog")

        self.top = self.xml.toplevel
        self.top.set_icon(ICON)
        self.top.set_title("%s - Gramps" % msg1)

        self.task1 = task1
        self.task2 = task2
        self.task3 = task3

        label1 = self.xml.get_object("label4")
        label1.set_text('<span weight="bold" size="larger">%s</span>' % msg1)
        label1.set_use_markup(True)

        label2 = self.xml.get_object("label3")
        label2.set_text(msg2)
        label2.set_use_markup(True)

        check_button = self.xml.get_object("use_always")

        if parent:
            self.top.set_transient_for(parent)
            parent_modal = parent.get_modal()
            if parent_modal:
                parent.set_modal(False)
        self.top.show()
        self.top.connect("delete_event", self.warn)
        response = Gtk.ResponseType.DELETE_EVENT

        # Need some magic here, because an attempt to close the dialog
        # with the X button not only emits the 'delete_event' signal
        # but also exits with the RESPONSE_DELETE_EVENT
        while response == Gtk.ResponseType.DELETE_EVENT:
            response = self.top.run()

        if response == 1:
            self.task1()
        elif response == 2:
            self.task2()
        elif response == 3:
            self.task3()
        if check_button.get_active():
            self.default_action = response
        else:
            self.default_action = 0
        self.top.destroy()
        if parent and parent_modal:
            parent.set_modal(True)

    def warn(self, obj, obj2):
        WarningDialog(
            _("Attempt to force closing the dialog"),
            _(
                "Please do not force closing this important dialog.\n"
                "Instead select one of the available options"
            ),
            parent=self.top,
        )
        return True


class MultiSelectDialog:
    """
    Allows a Yes, No, Cancel dialog that includes a checkbox
    with "Use this answer for the rest of the items" that works with 'Yes'
    """

    def __init__(
        self,
        msg1_func,
        msg2_func,
        items,
        lookup,
        cancel_func=None,
        no_func=None,
        yes_func=None,
        multi_yes_func=None,
        parent=None,
    ):
        """
        msg1_func a function to display big bold text at top
        msg2_func a function to display normal text at center of dialog
        items   a list of objects (often handles) passed to lookup
                Must be a list, not a generator
        lookup  A function that looks up an object to pass to yes/no/multi_func
        cancel_func     A function or None called on cancel.
        yes_func        A function or None called on Yes.
        no_func         A function or None called on No.
        multi_yes_func  A function or None called on Yes with
                        "use this answer for rest" checked.
        The above xxx_func calls will have parameters:
            object from lookup function
            "parent=" set to this dialog for transient parent purposes.
        """
        self.xml = Glade(toplevel="multiselectdialog")

        self.top = self.xml.toplevel
        self.top.set_icon(ICON)

        self.msg1_func = msg1_func
        self.msg2_func = msg2_func
        self.items = items
        self.lookup = lookup
        self.cancel_func = cancel_func
        self.no_func = no_func
        self.yes_func = yes_func

        label1 = self.xml.get_object("label6")
        label2 = self.xml.get_object("label5")
        check_button = self.xml.get_object("apply_to_rest")

        if parent:
            self.top.set_transient_for(parent)
            parent_modal = parent.get_modal()
            if parent_modal:
                parent.set_modal(False)
        self.top.connect("delete_event", self.warn)

        default_action = 0
        for indx, selected in enumerate(items):
            item = self.lookup(selected)
            if default_action == 0:
                msg1 = self.msg1_func(item)
                msg2 = self.msg2_func(item)

                self.top.set_title("%s - Gramps" % msg1)
                label1.set_text('<span weight="bold" size="larger">%s</span>' % msg1)
                label1.set_use_markup(True)
                label2.set_text(msg2)
                label2.set_use_markup(True)
                self.top.show()

                # Need some magic here, because an attempt to close the dialog
                # with the X button not only emits the 'delete_event' signal
                # but also exits with the RESPONSE_DELETE_EVENT
                response = Gtk.ResponseType.DELETE_EVENT
                while response == Gtk.ResponseType.DELETE_EVENT:
                    response = self.top.run()

                if check_button.get_active():
                    # run the multiple yes if 'do remainder' is checked
                    if multi_yes_func and response == 3:
                        multi_yes_func(items[indx:], parent=self.top)
                        break
                    default_action = response
            else:
                response = default_action
            # Now do it
            if response == 1:  # Cancel
                if self.cancel_func:
                    self.cancel_func(item, parent=self.top)
                break
            elif response == 2:  # No
                if self.no_func:
                    self.no_func(item, parent=self.top)
            elif response == 3:  # Yes
                if self.yes_func:
                    self.yes_func(item, parent=self.top)
        self.top.destroy()
        if parent and parent_modal:
            parent.set_modal(True)

    def warn(self, obj, obj2):
        WarningDialog(
            _("Attempt to force closing the dialog"),
            _(
                "Please do not force closing this important dialog.\n"
                "Instead select one of the available options"
            ),
            parent=self.top,
        )
        return True


class MessageHideDialog:
    def __init__(self, title, message, key, parent=None):
        self.xml = Glade(toplevel="hidedialog")

        self.top = self.xml.toplevel
        self.top.set_icon(ICON)
        self.top.set_title("%s - Gramps" % title)

        dont_show = self.xml.get_object("dont_show")
        dont_show.set_active(config.get(key))
        title_label = self.xml.get_object("title")
        title_label.set_text('<span size="larger" weight="bold">%s</span>' % title)
        title_label.set_use_markup(True)

        self.xml.get_object("message").set_text(message)

        dont_show.connect("toggled", self.update_checkbox, key)
        if parent:
            self.top.set_transient_for(parent)
            parent_modal = parent.get_modal()
            if parent_modal:
                parent.set_modal(False)
        self.top.run()
        self.top.destroy()
        if parent and parent_modal:
            parent.set_modal(True)

    def update_checkbox(self, obj, constant):
        config.set(constant, obj.get_active())
        config.save()


## Testing function of some of these dialogs
def main(args):
    win = Gtk.Window()
    win.set_title("Dialog test window")
    win.set_position(Gtk.WindowPosition.CENTER)
    # Set the mnemonic modifier on Macs to alt-ctrl so that it
    # doesn't interfere with the extended keyboard, see
    # https://gramps-project.org/bugs/view.php?id=6943
    if is_quartz():
        win.set_mnemonic_modifier(
            Gdk.ModifierType.CONTROL_MASK | Gdk.ModifierType.MOD1_MASK
        )

    def cb(window, event):
        Gtk.main_quit()

    win.connect("delete-event", cb)

    def test_info(obj):
        InfoDialog("The title", "This is a lot of info\n to show to all!", parent=win)

    vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
    win.add(vbox)

    btn1 = Gtk.Button("Info dialog")
    btn1.connect("clicked", test_info)
    vbox.pack_start(btn1, True, True, 0)

    win.show_all()
    Gtk.main()


if __name__ == "__main__":
    import sys

    # fall back to root logger for testing
    _LOG = logging.getLogger()
    sys.exit(main(sys.argv))
